# -*- coding: utf-8 -*-
"""
Created on Sun Jul 04 20:48:48 2021

@author: Theano Iliopoulou
Code citation: T. Iliopoulou, and D. Koutsoyiannis, PythOm: A python toolbox implementing recent advances in rainfall intensity (ombrian) curves, EGU General Assembly 2021, online, EGU21-389, doi:10.5194/egusphere-egu21-389, European Geosciences Union, 2021.
© Creative Commons Attribution 4.0 International (CC BY 4.0) 

"""


from __future__ import division
from math import *
import pandas as pd
import numpy as np
import datetime
from matplotlib import pyplot as plt
from matplotlib.ticker import FuncFormatter
import seaborn as sns
from scipy.optimize import minimize, differential_evolution, dual_annealing, fmin, fmin_powell, shgo


############# Step 0 : Data Preprocessing and calibration choices
### import daily series (optional)
rain_d=pd.read_excel('C:\\***********\\bologna_series.xlsx',sheet_name='daily') ## CHANGE directory according to user pc
start_date_d=datetime.date(rain_d['YEAR'][0],1,1)
end_date_d=datetime.date(rain_d['YEAR'][len(rain_d)-1],12,31)
idx_d = pd.date_range(start_date_d, end_date_d,freq='D')
curr_range_d=pd.to_datetime({'year':rain_d['YEAR'],'month':rain_d['MONTH'],'day':rain_d['DAY']},errors='coerce')
drain= pd.Series(rain_d['PRCP'],dtype='float64')
drain.index = curr_range_d
drain = drain.reindex(idx_d) 
    
### import hourly series
rain_h=pd.read_excel('C:\\***********\\bologna_series.xlsx',sheet_name='hourly') ## CHANGE directory according to user pc
start_date_h=rain_h['HOUR'][0]
end_date_h=rain_h['HOUR'][len(rain_h)-1]
idx_h = pd.date_range(start_date_h, end_date_h,freq='1h')
curr_range_h=rain_h['HOUR']
hrain= pd.Series(rain_h['PRCP'],dtype='float64')
hrain.index = curr_range_h
hrain = hrain.reindex(idx_h) 

### Choose calibration scales and define names of scales and scales' values in hours
# set names of scales as strings compatible with pandas datetime's frequencies (e.g. 1 h as '1H', 10 min as '10min', 5 days as '5D')
scales_h=['1H','2H','4H','6H','12H','24H','48H','96H']  # scales to estimate from hourly series (may overlap with scales from daily series)
scales_d=['1D','2D','4D','8D','16D','32D','64D','183D','365D','731D','1461D','2922D','5844D'] # scales to estimate from daily series (optional)

# set all unique scales in hours
k_h=np.array([1,2,4,6,12,24,48,96,24,48,96,192,384,768,1536,4392,8760,17544,35064,70128,140256])

h_len=len(hrain[hrain.notnull()]) ## length of valid hourly data 
d_len=len(drain[drain.notnull()]) ## length of valid daily data 

## pre-allocate intensities dataframe
scales=scales_h+scales_d

df = pd.DataFrame(np.empty([len(hrain),len(scales)]))
df[:] = np.NaN
df.columns=scales
    
### Estimate empirical intensities series for each scale of hourly series
for i in scales_h:
    intensity=hrain.groupby(pd.Grouper(freq=i)).mean()
    intensity=intensity/k_h[0]  ## to convert from basic scale to mm/h
    intensity=intensity.reset_index(drop=True)
    df[i]=df[i].combine_first(intensity)

### Estimate empirical intensities series for each scale of daily series (optional)   
for i in scales_d: 
    intensity=drain.groupby(pd.Grouper(freq=i)).mean()
    intensity=intensity/24 ## divide with 24 h to convert from mm/d to mm/h
    intensity=intensity.reset_index(drop=True)
    df[i]=df[i].combine_first(intensity)  

############# Step I: Choose climacogram model
var_k=df.var(axis=0)    # Empirical climacogram

# Obtain a first estimate of model parameters related to climacogram from climacogram fitting and evaluate climacogram types
x0_CD=[0.05, 2, 2, 0.7]
x0_C=[0.05, 2, 0.5, 0.7]

# parameter bounds for climacogram types CD and C
bndsCD=((0.00001, None), (0.00001, None),(0.00001,None),(0.00001,0.95))
bndsC=((0.00001, None), (0.00001, None),(0.00001,1),(0.00001,0.95))

# fit climacogram models
res_CD=minimize(CgCD_Error, x0_CD, args=(h_len,d_len,var_k,k_h,scales_h,scales_d), bounds=bndsCD, method='L-BFGS-B', options={'disp': True})
res_C=minimize(CgC_Error, x0_C, args=(h_len,d_len,var_k,k_h,scales_h,scales_d), bounds=bndsC, method='L-BFGS-B', options={'disp': True})

# choose climacogram type based on fit
if res_CD.fun<res_C.fun:
    print "best fit type CD"
    print "error:", str(res_CD.fun)
    print "parameters:", str(res_CD.x)
elif res_CD.fun>res_C.fun:
    print "best fit type C"
    print "error:", str(res_.fun)
    print "parameters:", str(res_C.x)  
            
############# Step II: probability wet/dry  structure            
nw_k=df.apply(lambda x: len(x[x>0]), axis=0)                       # length of wet days for each scale            
n_k=df.apply(lambda x: len(x[x.notnull()]), axis=0)                # number of scales with data
pw_k=df.apply(lambda x: len(x[x>0])/len(x[x.notnull()]), axis=0)   # Empirical probability wet for each scale          
 
## choose transition timescale in hours
k_trans=96 # default value of 24 h


#############  Step III:  High-order moment estimation using K-moments      
# sort wet days intensities in increasing order
df_wet=df.copy()
df_wet[df_wet==0]=np.NaN
df_wet_sorted= pd.DataFrame(np.sort(df_wet.values, axis=0), index=df_wet.index, columns=df_wet.columns)

     
## Estimate K-moments
## Uncentral unbiased K-moments 
n_p=80                                              # choose number of moment orders or use default (n_p=80)
j_p=np.arange(0,n_p+1,1)
p = pd.DataFrame(np.empty([n_p+1,len(scales)]))     # pre-allocate moment orders dataframe
p.columns=scales
p[:] = np.NaN
    
for i in scales:
    p[i][:]=nw_k[i]**(j_p[:]/n_p)                   # orders in geometric progression
        
kmoms = pd.DataFrame(np.empty([n_p+1,len(scales)])) # pre-allocate K-moments dataframe
kmoms.columns=scales
kmoms[:] = np.NaN
    
for i in scales:
    kmoms[i]=[kmom(j,df_wet_sorted[i].values) for j in p[i]] # K-moments for orders p
    

#############  Step V:  Calibration  
##   Using local minimimzers (quicker, less precise, need altering x_g)
# parameter bounds
bnds=((0.00001, 10), (0.00001, 1),(0.00001,100),(0.00001,100),(0.00001,100),(0.00001,0.95),(0.00001,0.4)) # For CD-type
#bnds=((0.00001, 10), (0.00001, 1),(0.00001,100),(0.00001,100),(0.00001,1),(0.00001,0.95),(0.00001,0.4)) # For C-type  

# Initialize x_g starting point for minimization
m_g=df.iloc[:,-1].mean()             # Mean intensity parameter (params[0])
th_g=0.5                             # Exponent of the expression of probability dry parameter (params[1])
l1_g,l2_g,a_g,H_g=res_CD.x           # For Climacogram CD type: Intensity scale parameters (params[2],params[3]), Time scale parameter (params[4]), Hurst parameter (params[5]) 
#l1_g,a_g,m_g,H_g=res_C.x            # For Climacogram C type: Intensity scale parameter(params[2]), Time scale parameter (params[3]), Fractal parameter (params[4]), Hurst parameter (params[5]) 
xi_g=0.15                            # Tail index parameter (params[6])
x_g=[m_g,th_g,l1_g,l2_g,a_g,H_g,xi_g]

# Or manually set up x_g 
#x_g=[0.07, 0.8, 0.0008, 2, 8.24, 0.95, 0.11]

# Nelder-Mead (Suggested, bounds inserted as penalty in ombrian_quant_error function)
res_nm=minimize(ombrian_quant_error, x_g, args=(h_len,d_len,k_trans,k_h,scales_h,scales_d,'CD',pw_k,kmoms,p), method='Nelder-Mead', bounds=bnds, tol=None, callback=None, options={ 'maxiter': None, 'maxfev': None, 'disp': True, 'return_all': False, 'initial_simplex': None, 'xatol': 0.0001, 'fatol': 0.0001, 'adaptive': False})

# TNC minimizer (Suggested)
res_tnc=minimize(ombrian_quant_error, x_g, args=(h_len,d_len,k_trans,k_h,scales_h,scales_d,'CD',pw_k,kmoms,p), bounds=bnds, method='TNC', options={'disp': True,'maxiter':40000})

##   Using global minimizers
# Differential evolution (Suggested)
res_de=differential_evolution(ombrian_quant_error, args=(h_len,d_len,k_trans,k_h,scales_h,scales_d,'CD',pw_k,kmoms,p), bounds=bnds) 


#############  Step VI:  Simulation and visualization of the results (optional)    
# Parameter set
#prmtrs=res_tnc.x   ##  from TNC algorithm 
#prmtrs=res_nm.x    ##  from Nelder-Mead algorithm 
#prmtrs=res_de.x    ##  from Differential Evolution algorithm 

# Best parameter results for Bologna obtained from differential_evolution minimization
prmtrs= [7.36233688e-02, 9.94586141e-01, 1.17837329e-03, 2.15145983e+00,8.55033253e+00, 9.50000000e-01, 1.12463513e-01]

#### Simulation for a range of return periods
# choose maximum return period 
max_rp=1000
k_y=np.array(k_h/(365.25*24))
sim_rp = pd.DataFrame(np.empty([2000,len(scales)])) # set matrix with 2000 theoretical return periods (RP_min=minimum scale, RP_max=1000)
sim_rp[:] = np.NaN
sim_rp.columns=scales
sim_rp.iloc[0,:]=k_y+0.0101   
for i in scales:
    sim_rp[i][1:]=np.logspace(np.log(sim_rp[i][0]),np.log(max_rp),1999,base=e)  ## space return periods evenly in log space   

sim_x = ombrian_model(prmtrs,k_trans,k_h,scales,sim_rp,'CD')
sim_x[sim_x<prmtrs[0]]=np.NaN

## Estimate empirical return periods based on K-moments 
K_rp=K_return_periods(prmtrs,h_len,d_len,k_trans,k_h,scales_h,scales_d,'CD',pw_k,p)


#### Plot

plot_scales=['1H','2H','4H','6H','12H','24H','48H','96H','1D','2D','4D','8D','16D','32D','64D','183D','365D','1461D','5844D']
names_scales=['1 h','2 h','4 h','6 h','12 h','1 d','2 d','4 d','8 d','16 d','32 d','64 d ','0.5 y','1 y','4 y','16 y'] # names of scales for annotation

off_d=3 ## range of overlapping scales from daily & hourly series to avoid double plotting of simulated series (and keep consistent coloring)

sns.set()
sns.set_style("ticks")
plt.subplots(1,1,figsize=(8,8))
plt.subplots_adjust(left=0.11,right=0.93,bottom=0.1,top=0.95)


p1=plt.subplot(1, 1,1)
ax = plt.gca()
ax.loglog()
plt.tick_params(axis = "x", which='both',top = False)  
plt.tick_params(axis = "y",which='both', right = False)  
       
for axis in [ax.xaxis, ax.yaxis]:
    formatter = FuncFormatter(lambda y, _: '{:.16g}'.format(y))
    axis.set_major_formatter(formatter)

plt.xticks(np.array([0.01,0.10,1,10,100,1000]))

ax.set_xlim([0.01,max_rp+20])
ax.set_ylim([0.01,100])

## add second palette if length>12
if len(names_scales)<=12:
   clrs=sns.color_palette("Paired", len(names_scales))
elif len(names_scales)>12:
   add_clrs=sns.color_palette('gist_stern_r',len(names_scales)-12)
   clrs=sns.color_palette("Paired", 12) + add_clrs

## empirical + simulation for hourly
for j in np.arange(0,len(scales_h)):
    scale=scales_h[j] 
    plt.scatter(K_rp[scale].values,kmoms[scale].values,s=30, linewidths=0.8,facecolors='none',edgecolor=clrs[j],label=None)
    plt.plot(sim_rp[scale].values,sim_x[scale].values,'-',color=clrs[j],linewidth=2)

## empirical for daily
for j in np.arange(len(scales_h),len(plot_scales)):
    scale=plot_scales[j] 
    plt.scatter(K_rp[scale].values,kmoms[scale].values,marker='D',s=25, facecolors=clrs[j-off_d],edgecolor=clrs[j-off_d],label=None)
    
## simulation for daily
for j in np.arange(len(scales_h),len(plot_scales)-off_d):
    scale=plot_scales[j+off_d] 
    plt.plot(sim_rp[scale].values,sim_x[scale].values,'-',color=clrs[j],linewidth=2)

props = dict(boxstyle='square', facecolor='gainsboro',edgecolor='gainsboro',linewidth=1, alpha=0.3)

## add second legend
plt.scatter(1000,1000,s=30, linewidths=0.8,facecolors='none',edgecolor='k',label='Empirical (K-moments) from hourly series')
plt.scatter(1000,1000,marker='D',s=25, facecolors='k',edgecolor='k',label='Empirical (K-moments) from daily series')
plt.plot(1000,1000,'-',color='k',linewidth=2, label='Ombrian model')

plt.ylabel("Precipitation intensity $x$ (mm/h)",fontsize=14);
#    
plt.xlabel("Return Period $T$ (years)",fontsize=14);

plt.legend(frameon=0,fancybox=1,bbox_to_anchor=(0.67, 1.02),loc=1,ncol=1,fontsize=14)  
plt.grid(b=True, which='major',axis='both', color='0.90',linestyle='-',zorder=2)
sns.despine(right=True)

## annotate names to curves
for j in np.arange(0,len(scales_h)):
    plt.annotate(names_scales[j],(max_rp+20,sim_x.iloc[:,j].max()),color=clrs[j],weight='bold',fontsize=12)

for j in np.arange(len(scales_h),len(plot_scales)-off_d):
    scale=plot_scales[j+off_d] 
    plt.annotate(names_scales[j],(max_rp+20,sim_x[scale].max()-0.006*j*sim_x[scale].max()),color=clrs[j],weight='bold',fontsize=12)

## save figure
plt.savefig("C:\\********\\ombrian_curves.png",dpi=300,facecolour='white') ## choose directory to save figure

plt.close(); 
         
